using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Data;
using System.Globalization;
using System.Reflection;
using System.Text;
using System.Threading;
using Epsilon.Screens.Interfaces;
using Epsilon.Screens.Interfaces.Definitions;
using Epsilon.Screens.Interfaces.ShowRoom;
using Produmex.Foundation;
using Produmex.Foundation.Barcode;
using Produmex.Foundation.Data.DbClient;
using Produmex.Foundation.Data.Sbo;
using Produmex.Foundation.Data.Sbo.BusinessObjects;
using Produmex.Foundation.Data.Sbo.BusinessObjects.Definitions;
using Produmex.Foundation.Data.Sbo.BusinessObjects.Definitions.Tables;
using Produmex.Foundation.Data.Sbo.DataObjects;
using Produmex.Foundation.Data.Sbo.Providers;
using Produmex.Foundation.Data.Sbo.Utilities;
using Produmex.Foundation.Data.SqlClient;
using Produmex.Foundation.Diagnostics;
using Produmex.Foundation.GS1;
using Produmex.Foundation.Messages;
using Produmex.Foundation.Reflection;
using Produmex.Foundation.SlimScreen;
using Produmex.Foundation.SlimScreen.Interfaces;
using Produmex.Foundation.SlimScreen.Interfaces.Definitions;
using Produmex.Foundation.SlimScreen.Interfaces.Definitions.DataObjects;
using Produmex.Foundation.SlimScreen.Interfaces.Definitions.KnownDataSets;
using Produmex.Foundation.SlimScreen.WinGui;
using Produmex.Foundation.Utilities;
using Produmex.Foundation.Workflows;
using Produmex.Foundation.Workflows.Parameters;
using Produmex.Foundation.Wwf.Sbo;
using Produmex.Foundation.Wwf.Sbo.LocalServices;
using Produmex.Sbo.Logex.Data.BusinessObjects;
using Produmex.Sbo.Logex.Data.BusinessObjects.Definitions;
using Produmex.Sbo.Logex.Data.BusinessObjects.Definitions.Tables;
using Produmex.Sbo.Logex.Data.DataObjects;
using Produmex.Sbo.Logex.Data.Providers;
using Produmex.Foundation.Data.DataSetClient;


// PmxWorkflowExecutionTypes.SUB_FLOW
namespace Produmex.Sbo.Logex.WorkflowScripts
{
    public class WorkflowScript_EPSILON_StockCountingFlowScript : WorkflowInstanceScriptBase
    {
        private static readonly ILog s_log = LogProvider.GetLogger(MethodInfo.GetCurrentMethod().DeclaringType);


        private static class Defs
        {
            /// <summary>
            /// NumberOfAllowedFailedAdminPassword
            /// </summary>
            public const int NumberOfAllowedFailedAdminPassword = 3;
        }



        // Input parameters
        public ReadOnlyBinder<CultureInfo> DefaultCultureInfo;
        public ReadOnlyBinder<PmxOseCompany> PmxOseCompany;



        private CultureInfo current_DefaultCultureInfo
        {
            get { return DefaultCultureInfo.Get(); }
            //set { DefaultCultureInfo.Set(value); }
        }

        private PmxOseCompany currentPmxOseCompany
        {
            get { return PmxOseCompany.Get(); }
            //set { PmxOseCompany.Set(value); }
        }

        // Input parameters
        // <none>

        // Output parameters
        // <none>


        // Menu options
        private class MenuOption
        {
            public string ButtonKey;
            public WorkflowInstanceBase Flow;

            public MenuOption(string buttonKey, WorkflowInstanceBase flow)
            {
                ButtonKey = buttonKey;
                Flow = flow;
            }
        };
        private List<MenuOption> m_MenuOptions = new List<MenuOption>();

        public WorkflowScript_EPSILON_StockCountingFlowScript(WorkflowInstanceBase parent, WorkflowInstanceFactory factory)
            : base(parent, factory)
        {


        }

        #region WorkflowInstanceScriptBase Members

        protected override void Execute()
        {

            // Parameters in scope
            Session session = GetScopeParameter("Session") as Session;

            ISboProviderService sboProviderService = GetScopeParameter("<WwfService>ISboService") as ISboProviderService;
            string deviceID = session.DeviceInstanceId;
            string initialErrorKey = null;

            Dictionary<string, PmxParameterValue> parameters = new Dictionary<string, PmxParameterValue>();
            string msgKey = null;
            ProdumexError pmxError = null;
            Message msg = null;
            string buttonKeySelectTask = null;
            bool isAdmin = false;
            DataSet dsCustomers = null;
            string query = null;
            string customerCode = null;
            string customerName = null;
            string adminPassword = null;
            int numberOfWrongEntriesAdminPassword = 0;
            BarcodeContainer barCode = null;
            string GTIN = null;
            string itemCode = null;
            DataSet dsItems = null;
            string itemDescription = null;
            PmxItemInfo itemInfo = null;
            bool isSelectProduct = false;
            string columnNameNumInSales = null;
            ProductInfo productInfo = null;
            double? customerPrice = null;
            double? itemCost = null;
            double numberOfItems = 1;
            string customerPriceCurrency = null;
            double availableQuantity = 0;
            ItemPrice itemPrice = null;
            Dictionary<string, PmxItemInfo> listOfItemInfos = new Dictionary<string, PmxItemInfo>();
            SboCompanyAdminInfo sboCompanyAdminInfo = null;
            double quantityToOrder = 0;
            double currentPrice = 0;
            double price = 0;
            string currency = null;
            string customerCurrency = null;
            DataSet dsSalesOrderLines = null;
            string buttonKeyIdentifyProduct = null;
            DateTime? deliveryDate = null;
            int? creditNoteDocEntry = null;
            int numberOfReports = 0;
            string buttonKeySelectTaskBeforeOrdering = null;
            bool askReprint = false;
            // double numberOfColliDouble = 0;
            // int numberOfColli = 0;
            string creditNoteRemark;
            string FolioNumberPart1;
            string FolioNumberPart2;

            string lineRemark = null;

            string correctionTaxCode = null;
            string correctionTaxCodeLine = null;


            try
            {
                //get admin info
                sboCompanyAdminInfo = sboProviderService.InvokeMethodWithDbConnection<SboCompanyAdminInfo>(true, false, null, null, new DelegateWithPmxDbConnection<SboCompanyAdminInfo>(ExecuteGetSboCompanyInfo));
                //get parameters
                parameters = sboProviderService.InvokeMethodWithDbConnection<Dictionary<string, PmxParameterValue>>(false, false, null, null, new DelegateWithPmxDbConnection<Dictionary<string, PmxParameterValue>>(ExecuteGetParametersForFlow), deviceID);
                adminPassword = parameters[PmxExtensionParameterCodes.SHOWROOM_ADMIN_PASSWORD].ParameterValue;
                columnNameNumInSales = parameters[PmxExtensionParameterCodes.SHOWROOM_NUM_IN_SALES_COLUMN_NAME].ParameterValue;

                if (string.IsNullOrEmpty(columnNameNumInSales))
                {
                    //set default
                    columnNameNumInSales = PmxItemTable.Columns.NumberOfItemsPerSalesUnit.NAME;
                }

                numberOfWrongEntriesAdminPassword = 0;
            }
            catch (ProdumexException ex)
            {
                msgKey = ex.Message;
                pmxError = ex.Error;
            }
            catch (Exception ex)
            {
                msgKey = ex.Message;
            }

            if (msgKey != null)
            {
                session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                    current_DefaultCultureInfo, BuildParamCollection(
                        "InitialErrorKey", null,
                        "MessageKey", msgKey,
                        "PmxError", pmxError,
                        "ShowButton", true));
                WaitForMessage();
                return;
            }



        Step_SelectTask:

            #region SELECT TASK

            buttonKeySelectTask = null;
            isAdmin = false;

            session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.ISelectButtonScreen),
            current_DefaultCultureInfo, BuildParamCollection(
                "InitialErrorKey", initialErrorKey,
                "TitleKey", "MSG_TITLE_SELECT_TASK",
                "AllowedButtons", new string[] { "MSG_BUTTON_SELECT_ITEM", "MSG_BUTTON_ENTER_ADMIN_PASSWORD" },
                "ShowStopButton", true
                ));
            msg = WaitForMessage();

            initialErrorKey = null;

            if (msg.Name.EndsWith(".ButtonClicked"))
            {
                buttonKeySelectTask = ExtractParameter<string>(msg.Parameters, "buttonKey");

                if (buttonKeySelectTask == "MSG_BUTTON_SELECT_ITEM")
                {
                    goto Step_IdentifyProduct;
                }
                else
                {
                    goto Step_EnterAdminPassword;
                }

            }

            return;


            #endregion

        Step_EnterAdminPassword:

            #region ENTER ADMIN PASSWORD

            isAdmin = false;

            session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IEnterStringValueScreen),
            current_DefaultCultureInfo, BuildParamCollection(
                "InitialErrorKey", initialErrorKey,
                "TitleKey", "MSG_TITLE_ENTER_ADMIN_PASSWORD",
                "AllowToGoBack", true
                ));
            msg = WaitForMessage();

            initialErrorKey = null;

            if (msg.Name.EndsWith(".StringEntered"))
            {
                string tempPassword = ExtractParameter<string>(msg.Parameters, "stringValue");
                if (string.Compare(tempPassword, adminPassword, false) != 0)
                {
                    if (Defs.NumberOfAllowedFailedAdminPassword == numberOfWrongEntriesAdminPassword)
                    {
                        //log off, because number of fails has been reached
                        session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                          current_DefaultCultureInfo, BuildParamCollection(
                        "InitialErrorKey", null,
                        "MessageKey", "MSG_ERROR_WRONG_ADMIN_PASSWORD_NO_MORE_RETRIES",
                        "ShowButton", true));
                        WaitForMessage();


                        return;
                    }
                    //wrong pasword!
                    initialErrorKey = "MSG_ERROR_WRONG_ADMIN_PASSWORD";
                    numberOfWrongEntriesAdminPassword++;
                    goto Step_EnterAdminPassword;
                }
                else
                {
                    //correct password
                    numberOfWrongEntriesAdminPassword = 0;
                    isAdmin = true;
                    goto Step_SelectCustomer;
                }
            }

            //back requested
            goto Step_SelectTask;

            #endregion

        Step_SelectCustomer:

            #region SELECT CUSTOMER

            customerCode = null;
            customerName = null;
            correctionTaxCode = null;
            query = BuildQuery.GetCustomers();
            dsCustomers = sboProviderService.RunView(false, null, null, query);

            session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.ISelectCustomerScreen),
            current_DefaultCultureInfo, BuildParamCollection(
                "InitialErrorKey", initialErrorKey,
                "TitleKey", "MSG_TITLE_SELECT_CUSTOMER",
                "CustomerDS", dsCustomers
                ));
            msg = WaitForMessage();

            initialErrorKey = null;

            if (msg.Name.EndsWith(".CustomerSelected"))
            {
                customerCode = ExtractParameter<string>(msg.Parameters, "customerCode");

                DataRow[] customerRows = dsCustomers.Tables[0].Select(CustomerDataSet.Columns.CUSTOMER_CODE.Name + "='" + customerCode + "'");
                if (customerRows.Length < 1)
                {
                    initialErrorKey = KnownMessageKeys.MSG_ERROR_OPERATION_NOT_SUPPORTED;
                    goto Step_SelectTask;
                }

                customerName = Convert.ToString(customerRows[0][CustomerDataSet.Columns.CUSTOMER_NAME.Name]);
                customerCurrency = Convert.ToString(customerRows[0][SboBusinessPartnerTable.Columns.Currency.NAME]);
                correctionTaxCode = Convert.ToString(customerRows[0]["VatCrctn"]);

                //new sales order dataset
                dsSalesOrderLines = MakeNewSalesOrderDataSet();

                goto Step_IdentifyProduct;

            }

            //back requested
            goto Step_SelectTask;

            #endregion

        Step_IdentifyProduct:

            #region IDENTIFY PRODUCT

            lineRemark = null;
            buttonKeyIdentifyProduct = null;
            List<string> allowedButtons = new List<string>();

            if (!currentPmxOseCompany.DisableItemSelectionInFlows)
            {
                allowedButtons.Add("MSG_BUTTON_SELECT_PRODUCT");
            }

            if (this.GetNumberOfRows(dsSalesOrderLines) > 0)
            {
                allowedButtons.Add("MSG_BUTTON_MAKE_CREDIT_NOTE_EPSILON");
            }


            session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IIdentifyGS1DataScreen),
            current_DefaultCultureInfo, BuildParamCollection(
                "InitialErrorKey", initialErrorKey,
                "TitleKey", "MSG_TITLE_IDENTIFY_PRODUCT",
                "AllowedButtons", allowedButtons.Count > 0 ? allowedButtons.ToArray() : null,
                "ImageType", GS1ImageType.BoxImage
                ));
            msg = WaitForMessage();

            //clear data
            initialErrorKey = null;
            barCode = null;
            GTIN = null;
            itemDescription = null;
            isSelectProduct = false;
            correctionTaxCodeLine = null;

            if (msg.Name.EndsWith(".DataEntered"))
            {
                barCode = ExtractParameter<BarcodeContainer>(msg.Parameters, "data");

                //check if there is a product present
                if (barCode != null &&
                !string.IsNullOrEmpty(BarcodeHelper.GetProductIdentification(barCode)))
                {
                    GTIN = BarcodeHelper.GetProductIdentification(barCode);
                    goto Step_SelectProduct;
                }
                else
                {
                    initialErrorKey = "MSG_ERROR_NO_PRODUCT_ON_BARCODE";
                    goto Step_IdentifyProduct;
                }

            }

            if (msg.Name.EndsWith(".ButtonClicked"))
            {
                buttonKeyIdentifyProduct = ExtractParameter<string>(msg.Parameters, "buttonKey");

                if (buttonKeyIdentifyProduct == "MSG_BUTTON_SELECT_PRODUCT")
                {
                    //select the product
                    isSelectProduct = true;
                    goto Step_SelectProduct;
                }
                else
                {
                    goto Step_ShowAllOrderedItems;
                }
            }


            //if we come here, we want to go back.
            if (GetNumberOfRows(dsSalesOrderLines) == 0)
            {
                goto Step_SelectTask;
            }
            else
            {
                //there are ordered items, ask user if he wants to go back
                session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IDecisionScreen),
                    current_DefaultCultureInfo, BuildParamCollection(
                        "MessageKey", "MSG_QUESTION_ORDERED_ITEMS_IN_LIST",
                        "TitleKey", "MSG_TITLE_QUESTION_ORDERED_ITEMS_IN_LIST"
                        ));
                msg = WaitForMessage();

                if (msg.Name.EndsWith(".Yes"))
                {
                    //go back
                    goto Step_SelectTask;
                }
                else
                {
                    //go back
                    goto Step_IdentifyProduct;
                }
            }


            #endregion

        Step_SelectProduct:

            #region SELECT PRODUCT


            //no items in list, so run the query again
            query = BuildQuery.GetProductsAndSpecialPrices(GTIN, null, Produmex.Sbo.Logex.Data.SboCommands.PackagingUnit.Sales, customerCode, columnNameNumInSales, true, sboProviderService.GetDbTool());
            dsItems = sboProviderService.RunView(false, null, null, query);


            if (this.GetNumberOfRows(dsItems) == 1 )//&&        !isSelectProduct)
            {
                //there is only one item, so auto select it!
                DataRow row = dsItems.Tables[0].Rows[0];
                //copy the data
                itemDescription = row[ProductDataSet.Columns.PRODUCT_DESCRIPTION.Name] as string;
                itemCode = row[ProductDataSet.Columns.PRODUCT_CODE.Name] as string;

                itemCost = Convert.ToDouble(row[SboItemTable.Columns.ItemCost.NAME]);

                // //   double tempPrice;
                // //   if (double.TryParse(Convert.ToString(row[SboSpecialPricesTable.Columns.Price.NAME]), out tempPrice))
                // //   {
                // //       customerPrice = tempPrice;
                ////    }
                //    else
                //    {
                //        customerPrice = null;
                //    }
                //customerPriceCurrency = row[SboSpecialPricesTable.Columns.Currency.NAME] as string;

                availableQuantity = Convert.ToDouble(row[PmxInventoryTotalTable.Columns.Quantity.NAME]);

                numberOfItems = Convert.ToInt32(row[columnNameNumInSales]);

                //correctionTaxCodeLine = row["VatCrctn"] as string;


                //show info Not needed, directly go to Entery Quantity
                // goto Step_ShowItemInfo;

                goto Step_EnterQuantity;
            }

            session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.ISelectProductScreen),
                current_DefaultCultureInfo, BuildParamCollection(
                    "InitialErrorKey", initialErrorKey,
                    "TitleKey", "MSG_TITLE_SELECT_PRODUCT",
                    "ProductDS", dsItems
                    ));
            msg = WaitForMessage();

            //clear data
            initialErrorKey = null;

            if (msg.Name.EndsWith(".ProductSelected"))
            {
                itemCode = ExtractParameter<string>(msg.Parameters, "itemCode");
                //get itemdescription
                DataRow[] rows = dsItems.Tables[0].Select(string.Format("{0} = '{1}'", ProductDataSet.Columns.PRODUCT_CODE.Name, itemCode));
                if (rows.Length > 0)
                {
                    //copy the data
                    itemDescription = rows[0][ProductDataSet.Columns.PRODUCT_DESCRIPTION.Name] as string;

                    itemCost = Convert.ToDouble(rows[0][SboItemTable.Columns.ItemCost.NAME]);

                    double tempPrice;
                    //if (double.TryParse(Convert.ToString(rows[0][SboSpecialPricesTable.Columns.Price.NAME]), out tempPrice))
                    //{
                    //    customerPrice = tempPrice;
                    //}
                    //else
                    //{
                    //    customerPrice = null;
                    //}
                    //customerPriceCurrency = rows[0][SboSpecialPricesTable.Columns.Currency.NAME] as string;

                    availableQuantity = Convert.ToDouble(rows[0][PmxInventoryTotalTable.Columns.Quantity.NAME]);

                    numberOfItems = Convert.ToInt32(rows[0][columnNameNumInSales]);

                    // correctionTaxCodeLine = rows[0]["VatCrctn"] as string;


                }
                //show info
                goto Step_ShowItemInfo;
            }

            goto Step_IdentifyProduct;

            #endregion

        Step_ShowItemInfo:

            #region SHOW ITEM INFO

            if (!listOfItemInfos.ContainsKey(itemCode))
            {
                listOfItemInfos.Add(itemCode, sboProviderService.InvokeMethodWithDbConnection<PmxItemInfo>(false, false, null, null, new DelegateWithPmxDbConnection<PmxItemInfo>(ExecuteGetItemInfo), itemCode));
            }

            //get the iteminfo
            itemInfo = listOfItemInfos[itemCode];
            // itemPrice = sboProviderService.InvokeMethodWithDbConnection<ItemPrice>(true, false, null, null, new DelegateWithPmxDbConnection<ItemPrice>(ExecuteGetItemPrice), customerCode, itemCode, 1, DateTime.Now);

            //  price = itemPrice == null ? 0 : itemPrice.Price;
            //  currency = itemPrice == null ? customerCurrency : itemPrice.Currency;


            productInfo = new ProductInfo();
            productInfo.ItemCode = itemInfo.ItemCode;
            productInfo.ItemDescription = itemInfo.Description;
            productInfo.InventoryUom = itemInfo.InventoryUom;

            //productInfo.Price = price;
            //productInfo.Currency = currency;

            // productInfo.CustomerPrice = customerPrice;
            // productInfo.CustomerCurrency = customerPriceCurrency;

            if (isAdmin)
            {
                productInfo.ItemCost = itemCost;
                productInfo.ItemCostCurrency = sboCompanyAdminInfo.LocalCurrency;
            }
            else
            {
                productInfo.ItemCost = null;
            }


            productInfo.NumberOfItems = numberOfItems;
            productInfo.AvailableQuantity = availableQuantity;
            //Change by TBA


            //  session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.ShowRoom.IShowProductInfoScreen),
            //  current_DefaultCultureInfo, BuildParamCollection(
            //      "InitialErrorKey", initialErrorKey,
            //      "TitleKey", "MSG_TITLE_PRODUCT_INFO",
            //      "AllowedButtons", new string[1] { "MSG_BUTTON_OK" },
            //      "ProductInfo", productInfo
            //      ));
            //  msg = WaitForMessage();

            //clear data
            initialErrorKey = null;


            //   if (msg.Name.EndsWith(".ButtonClicked"))
            //  {
            //      goto Step_EnterQuantity;
            //  }

            //goto Step_IdentifyProduct;
            goto Step_EnterQuantity;


            #endregion

        Step_EnterQuantity:

            #region ENTER QUANTITY

            quantityToOrder = 0;

            lineRemark = null;

            // session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IEnterQuantityWithProductPricesScreen),
            session.ShowScreen(typeof(IEnterQuantityScreen),
                current_DefaultCultureInfo, BuildParamCollection(
               "InitialErrorKey", initialErrorKey,
               "TitleKey", "MSG_TITLE_ENTER_QUANTITY_TO_RETURN_EPSILON",
               "ProductInfo", productInfo,
               "InitialQuantity", 1,
               "NumberOfDigitsAfterDecimal", itemInfo.UomDecimals,
               "MaximumQuantity", -1,
               "UOM", itemInfo.InventoryUom,
               "InnerQuantity", itemInfo.NumberOfItemsPerSalesUnit,
               "InnerUOM", itemInfo.SalesUom,
               "ItemBarCode", itemInfo.CodeBars
               ));

            msg = WaitForMessage();

            //Change by TBA: we don't want to see the enter quantity screen, so we default it to 1
            if (msg.Name.EndsWith(".QuantityEntered"))
            {
                quantityToOrder = ExtractParameter<double>(msg.Parameters, "quantity");
                // quantityToOrder = 1;






                sboProviderService.InvokeMethodWithDbConnection<object>(false, false, null, null,
                          delegate(PmxDbConnection dbConn, object[] parameters2)
                          {
                              StringBuilder queryBuilder = new StringBuilder();

                              queryBuilder.AppendFormat("INSERT INTO {0} ({1},{2}) ",
                                    SqlCommandHelper.EscapeTableName("PMX_STOCKCOUNTING"),
                                    SqlCommandHelper.EscapeColumnName("ItemCode"),
                                    SqlCommandHelper.EscapeColumnName("Quantity"));

                              queryBuilder.AppendFormat(" VALUES ({0}, {1}) ",
                                   SqlCommandHelper.FormatAndEscapeColumnValue(itemCode, dbConn.DbCultureInfo),
                                   SqlCommandHelper.FormatAndEscapeColumnValue(quantityToOrder.ToString(), dbConn.DbCultureInfo));




                              SboRecordsetHelper.RunQueryNoReturn(s_log, queryBuilder.ToString(), dbConn);


                              return null;

                          });


            }

            goto Step_IdentifyProduct;

            #endregion

        Step_EnterPrice:

            #region ENTER PRICE

            currentPrice = price;

            //if (isAdmin || currentPrice == 0)
            //{
            //    //Only if admin is logged on, or price is 0, the user is allowed to change price

            //    session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IEnterQuantityWithProductPricesScreen),
            //        current_DefaultCultureInfo, BuildParamCollection(
            //       "InitialErrorKey", initialErrorKey,
            //       "TitleKey", "MSG_TITLE_ENTER_UNIT_PRICE",
            //       "ProductInfo", productInfo,
            //       "InitialQuantity", currentPrice,
            //       "NumberOfDigitsAfterDecimal", 2,
            //       "MaximumQuantity", -1,
            //       "UOM", customerCurrency
            //       ));
            //    msg = WaitForMessage();


            //    if (msg.Name.EndsWith(".QuantityEntered"))
            //    {
            //        currentPrice = ExtractParameter<double>(msg.Parameters, "quantity");
            //        goto Step_AddToTempDelivery;
            //    }

            //    //Go back
            //    goto Step_EnterQuantity;

            //}

            goto Step_AddToTempDelivery;

            #endregion

        Step_AddToTempDelivery:

            #region ADD TO TEMP DELIVERY

            DataRow newRow = dsSalesOrderLines.Tables[0].NewRow();

            int newKey = 0;

            foreach (DataRow row in dsSalesOrderLines.Tables[0].Rows)
            {
                if (Convert.ToInt32(row[SalesOrderLineEpsilonDataSet.Columns.INTERNAL_KEY.Name]) > newKey)
                {
                    newKey = Convert.ToInt32(row[SalesOrderLineEpsilonDataSet.Columns.INTERNAL_KEY.Name]);
                }
            }

            dsSalesOrderLines.Tables[0].Rows.Add(newRow);
            newKey++;

            newRow[SalesOrderLineEpsilonDataSet.Columns.INTERNAL_KEY.Name] = newKey;
            newRow[SalesOrderLineEpsilonDataSet.Columns.ITEMCODE.Name] = itemCode;
            newRow[SalesOrderLineEpsilonDataSet.Columns.ITEMDESCRIPTION.Name] = itemDescription;


            newRow[SalesOrderLineEpsilonDataSet.Columns.PRICE.Name] = currentPrice;
            newRow[SalesOrderLineEpsilonDataSet.Columns.QUANTITY.Name] = quantityToOrder;

            newRow[SalesOrderLineEpsilonDataSet.Columns.REMARKS.Name] = lineRemark;
            //  newRow[SalesOrderLineEpsilonDataSet.Columns.CORRECTION_VAT_CODE.Name] = string.IsNullOrEmpty(correctionTaxCodeLine) ? correctionTaxCode : correctionTaxCodeLine;
            newRow[SalesOrderLineEpsilonDataSet.Columns.CORRECTION_VAT_CODE.Name] = string.IsNullOrEmpty(correctionTaxCode) ? correctionTaxCodeLine : correctionTaxCode;
            newRow[SalesOrderLineEpsilonDataSet.Columns.CURRENCY.Name] = customerCurrency;
            newRow[SalesOrderLineEpsilonDataSet.Columns.UOM.Name] = string.IsNullOrEmpty(itemInfo.SalesUom) ? itemInfo.InventoryUom : itemInfo.SalesUom;


            goto Step_IdentifyProduct;


            #endregion

        Step_ShowAllOrderedItems:

            #region SHOW ALL ORDERED ITEMS

            //Group sales similar lines
            DataSet dsForScreen = dsSalesOrderLines.Clone();

            foreach (DataRow row in dsSalesOrderLines.Tables[0].Rows)
            {
                DataRow[] rows = dsForScreen.Tables[0].Select(string.Format("{0} = '{1}' AND {2} = {3}",
                    SalesOrderLineEpsilonDataSet.Columns.ITEMCODE.Name, row[SalesOrderLineEpsilonDataSet.Columns.ITEMCODE.Name],
                    SalesOrderLineEpsilonDataSet.Columns.PRICE.Name, Convert.ToString(row[SalesOrderLineEpsilonDataSet.Columns.PRICE.Name], CultureInfo.InvariantCulture)));

                if (rows.Length > 0)
                {
                    //add quantity
                    rows[0][SalesOrderLineEpsilonDataSet.Columns.QUANTITY.Name] = Convert.ToDouble(rows[0][SalesOrderLineEpsilonDataSet.Columns.QUANTITY.Name]) + Convert.ToDouble(row[SalesOrderLineEpsilonDataSet.Columns.QUANTITY.Name]);
                }
                else
                {
                    //Add row to DS
                    dsForScreen.Tables[0].ImportRow(row);
                }
            }


            session.ShowScreen(typeof(IShowAllOrderedItemsEpsilonScreen),
                current_DefaultCultureInfo, BuildParamCollection(
               "TitleKey", "MSG_TITLE_ALL_ITEMS_TO_RETURN_EPSILON",
               "OrderedItemsDS", dsForScreen
               ));
            msg = WaitForMessage();


            if (msg.Name.EndsWith(".ForwardRequested"))
            {
                dsSalesOrderLines = ExtractParameter<DataSet>(msg.Parameters, "dsOrderedItems");
                //order items
                goto Step_SelectTaskBeforeOrdering;
            }
            else
            {
                //go back
                goto Step_IdentifyProduct;
            }




            #endregion

        Step_SelectTaskBeforeOrdering:

            #region SELECT TASK BEFORE ORDERING

            buttonKeySelectTaskBeforeOrdering = null;

            session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.ISelectButtonScreen),
            current_DefaultCultureInfo, BuildParamCollection(
                "InitialErrorKey", initialErrorKey,
                "TitleKey", "MSG_TITLE_SELECT_TASK",
                "AllowedButtons", new string[] { "MSG_BUTTON_RETURN_EPSILON", "MSG_BUTTON_CANCEL_RETURN_EPSILON", "MSG_BUTTON_ADD_MORE_ITEMS" },
                "ShowStopButton", false
                ));
            msg = WaitForMessage();

            initialErrorKey = null;

            if (msg.Name.EndsWith(".ButtonClicked"))
            {
                buttonKeySelectTaskBeforeOrdering = ExtractParameter<string>(msg.Parameters, "buttonKey");

                if (buttonKeySelectTaskBeforeOrdering == "MSG_BUTTON_RETURN_EPSILON")
                {
                    goto Step_EnterDeliveryDate;
                }
                else if (buttonKeySelectTaskBeforeOrdering == "MSG_BUTTON_CANCEL_RETURN_EPSILON")
                {
                    dsSalesOrderLines.Clear();
                    goto Step_SelectTask;
                }
                else
                {
                    goto Step_IdentifyProduct;
                }

            }

            //go back
            goto Step_ShowAllOrderedItems;


            #endregion

        Step_EnterDeliveryDate:

            #region ENTER DELIVERY DATE

            deliveryDate = null;

            //ask for best before date
            /*    session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IEnterDateScreen),
                 current_DefaultCultureInfo, BuildParamCollection(
                     "InitialErrorKey", initialErrorKey,
                     "TitleKey", "MSG_TITLE_ENTER_DELIVERY_DATE",
                     "DefaultDate", DateTime.Now,
                     "AllowToGoBack", true
                     ));
                msg = WaitForMessage();
                */
            // reset error
            initialErrorKey = null;

            /*  if (msg.Name.EndsWith(".DateEntered"))
              {
                  deliveryDate = DateTime.Now;
                    //  ExtractParameter<DateTime?>(msg.Parameters, "date");

              
              }


              goto Step_ShowAllOrderedItems;
              goto Step_EnterNumberOfColli;*/

            deliveryDate = DateTime.Now;
            goto Step_EnterFolioNumberPart1;
        // goto Step_OrderItems;

            #endregion

        Step_OrderItems:

            #region ORDER ITEMS

            creditNoteDocEntry = null;

            try
            {
                // Inform the user we are saving the order
                session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                    current_DefaultCultureInfo, BuildParamCollection(
                        "MessageKey", "MSG_SAVING_SALES_RETURN_EPSILON",
                        "NoTranslationOfMessageKey", false,
                        "ShowButton", false));

                //make sales order
                creditNoteDocEntry = sboProviderService.InvokeMethodWithDbConnection<int?>(true, false, null, null, new DelegateWithPmxDbConnection<int?>(ExecuteMakeCreditNote),
                    customerCode, deliveryDate.Value, dsSalesOrderLines, FolioNumberPart1, FolioNumberPart2);

            }
            catch (Exception ex)
            {
                if (!(ex is ProdumexException))
                {
                    // Inform the user with the error
                    session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                        current_DefaultCultureInfo, BuildParamCollection(
                            "MessageKey", ExceptionHelper.Format(ExceptionHelper.Content.InnerExceptions, ex, current_DefaultCultureInfo),
                            "NoTranslationOfMessageKey", true,
                            "IsError", true,
                            "ShowButton", true));
                    WaitForMessage();
                }
                else
                {
                    // Inform the user with the error
                    session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                        current_DefaultCultureInfo, BuildParamCollection(
                            "PmxError", ((ProdumexException)ex).Error,
                            "ShowButton", true));
                    WaitForMessage();

                }
                //error occured, so go back
                goto Step_ShowAllOrderedItems;
            }

            //  goto Step_AskPrintDocuments;
            goto Step_AskSameCustomer;


            #endregion

        Step_PrintDocument:

            #region PRINT DOCUMENT

            numberOfReports = 0;

            if (creditNoteDocEntry.HasValue)
            {
                //only print if we have a doc entry
                try
                {
                    //trigger event
                    numberOfReports += sboProviderService.InvokeMethodWithDbConnection<int>(false, false, null, null,
                        delegate(PmxDbConnection conn, object[] parameters2)
                        {
                            PmxPrintReportProvider printProv = new PmxPrintReportProvider(conn);
                            return printProv.PrintReportsForEvent(PmxPrintReportEventType.SalesOrderHasBeenAdded,
                            null, deviceID, creditNoteDocEntry.Value, out askReprint);
                        });

                }

                catch (Produmex.Foundation.ProdumexException ex)
                {
                    s_log.Error(ex.ToString());
                    session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                        current_DefaultCultureInfo, BuildParamCollection(
                            "InitialErrorKey", initialErrorKey,
                            "MessageKey", KnownMessageKeys.MSG_ERROR_INTERNAL_ERROR,
                            "MessageAddendum", ex.Message,
                            "PmxError", ex.Error,
                            "ShowButton", true));

                    msg = WaitForMessage();
                }
                catch (Exception ex)
                {
                    s_log.Error(ex.ToString());
                    session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                        current_DefaultCultureInfo, BuildParamCollection(
                            "InitialErrorKey", initialErrorKey,
                            "MessageKey", KnownMessageKeys.MSG_ERROR_INTERNAL_ERROR,
                            "MessageAddendum", ex.Message,
                            "PmxError", Produmex.Sbo.Logex.Data.ProdumexErrors.UnexpectedError(s_log, ex),
                            "ShowButton", true));

                    msg = WaitForMessage();
                }


                if (numberOfReports > 0)
                {
                    goto Step_DocumentsPrinted;
                }
                else
                {
                    //Made sales order items
                    session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IShowMessageScreen),
                       current_DefaultCultureInfo, BuildParamCollection(
                           "InitialErrorKey", initialErrorKey,
                           "MessageKey", "MSG_SALES_ORDER_HAS_BEEN_MADE",
                           "ShowButton", true));

                    this.WaitForMessage();

                    goto Step_SelectTask;
                }

            }



            #endregion

        Step_DocumentsPrinted:

            #region DOCUMENTS PRINTED

            if (numberOfReports > 0 && askReprint)
            {
                session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IDecisionScreen),
                    current_DefaultCultureInfo, BuildParamCollection(
                        "MessageKey", "MSG_QUESTION_REPRINT_DOCUMENT",
                        "TitleKey", "MSG_TITLE_REPRINT_DOCUMENT"));

                msg = WaitForMessage();

                if (msg.Name.EndsWith(".Yes"))
                {
                    goto Step_PrintDocument;
                }

            }

            goto Step_SelectTask;

            #endregion





        Step_AskPrintDocuments:

            #region ASK PRINT DOCUMENTS

        //  if (salesDeliveryDocEntry.HasValue)
        //   {
        //       session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IDecisionScreen),
        //           current_DefaultCultureInfo, BuildParamCollection(
        //               "MessageKey", "MSG_QUESTION_PRINT_DOCUMENT",
        //               "TitleKey", "MSG_TITLE_PRINT_DOCUMENT"));

            //       msg = WaitForMessage();

            //     if (msg.Name.EndsWith(".Yes"))
        //       {
        //           goto Step_PrintDocument;
        //       }

            //  }

           // goto Step_SelectTask;


            #endregion


        Step_EnterNumberOfColli:

            #region ENTER NUMBER OF COLLI




            session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IEnterStringValueScreen),
                 current_DefaultCultureInfo, BuildParamCollection(
               "InitialErrorKey", initialErrorKey,
               "TitleKey", "MSG_TITLE_ENTER_NUMBER_OF_COLLI",
                //"InitialQuantity", 1,
                //"NumberOfDigitsAfterDecimal", 0,
                // "MaximumQuantity", -1
                 "AllowMultiLine", true
               ));
            msg = WaitForMessage();

            initialErrorKey = null;


            if (msg.Name.EndsWith(".StringEntered"))
            {
                // numberOfColliDouble = ExtractParameter<double>(msg.Parameters, "quantity");
                creditNoteRemark = ExtractParameter<string>(msg.Parameters, "stringValue");
                goto Step_OrderItems;
            }

            //not allowed to go back
            initialErrorKey = "MSG_ERROR_CANNOT_GO_BACK";
            goto Step_EnterNumberOfColli;



            #endregion


        Step_EnterFolioNumberPart1:

            #region ENTER NUMBER OF FOLIO PART 1




            session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IEnterStringValueScreen),
                 current_DefaultCultureInfo, BuildParamCollection(
               "InitialErrorKey", initialErrorKey,
               "TitleKey", "MSG_TITLE_ENTER_FOLIONUMBERPART1",
                //"InitialQuantity", 1,
                //"NumberOfDigitsAfterDecimal", 0,
                // "MaximumQuantity", -1
                 "AllowMultiLine", false

               ));
            msg = WaitForMessage();

            initialErrorKey = null;


            if (msg.Name.EndsWith(".StringEntered"))
            {
                // numberOfColliDouble = ExtractParameter<double>(msg.Parameters, "quantity");
                FolioNumberPart1 = ExtractParameter<string>(msg.Parameters, "stringValue");
                goto Step_EnterFolioNumberPart2;
            }

            //not allowed to go back
            initialErrorKey = "MSG_ERROR_CANNOT_GO_BACK";
            goto Step_EnterFolioNumberPart1;



            #endregion




        Step_EnterFolioNumberPart2:

            #region ENTER NUMBER OF FOLIO Part 2




            session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IEnterStringValueScreen),
                 current_DefaultCultureInfo, BuildParamCollection(
               "InitialErrorKey", initialErrorKey,
               "TitleKey", "MSG_TITLE_ENTER_FOLIONUMBERPART2",
                //"InitialQuantity", 1,
                //"NumberOfDigitsAfterDecimal", 0,
                // "MaximumQuantity", -1
                 "AllowMultiLine", false
               ));
            msg = WaitForMessage();

            initialErrorKey = null;


            if (msg.Name.EndsWith(".StringEntered"))
            {
                // numberOfColliDouble = ExtractParameter<double>(msg.Parameters, "quantity");
                FolioNumberPart2 = ExtractParameter<string>(msg.Parameters, "stringValue");
                goto Step_OrderItems;
            }

            //not allowed to go back
            initialErrorKey = "MSG_ERROR_CANNOT_GO_BACK";
            goto Step_EnterFolioNumberPart2;



            #endregion




        Step_AskSameCustomer:

            #region ASK SAME CUSTOMER

            if (creditNoteDocEntry.HasValue)
            {
                session.ShowScreen(typeof(Produmex.Foundation.SlimScreen.Interfaces.IDecisionScreen),
                    current_DefaultCultureInfo, BuildParamCollection(
                       "MessageKey", "MSG_CREDIT_NOTE_SAME_CUSTOMER",
                        "TitleKey", customerName));
                //   "TitleKey", "MSG_TITLE_CREDIT_NOTE_SAME_CUSTOMER"));


                msg = WaitForMessage();

                if (msg.Name.EndsWith(".Yes"))
                {
                    dsSalesOrderLines = MakeNewSalesOrderDataSet();
                    goto Step_IdentifyProduct;
                }

            }

            goto Step_SelectTask;


            #endregion


        }






        #endregion

        #region HELPERS

        private static DataSet MakeNewSalesOrderDataSet()
        {
            DataSet ds = new DataSet();

            ds.Tables.Add(SalesOrderLineEpsilonDataSet.NAME);
            ds.Tables[0].Columns.Add(SalesOrderLineEpsilonDataSet.Columns.CURRENCY.Name, typeof(string));
            ds.Tables[0].Columns.Add(SalesOrderLineEpsilonDataSet.Columns.INTERNAL_KEY.Name, typeof(int));
            ds.Tables[0].Columns.Add(SalesOrderLineEpsilonDataSet.Columns.ITEMCODE.Name, typeof(string));
            ds.Tables[0].Columns.Add(SalesOrderLineEpsilonDataSet.Columns.ITEMDESCRIPTION.Name, typeof(string));
            ds.Tables[0].Columns.Add(SalesOrderLineEpsilonDataSet.Columns.PRICE.Name, typeof(double));
            ds.Tables[0].Columns.Add(SalesOrderLineEpsilonDataSet.Columns.QUANTITY.Name, typeof(double));
            ds.Tables[0].Columns.Add(SalesOrderLineEpsilonDataSet.Columns.UOM.Name, typeof(string));
            ds.Tables[0].Columns.Add(SalesOrderLineEpsilonDataSet.Columns.NORMAL_PRICE.Name, typeof(double));

            ds.Tables[0].Columns.Add(SalesOrderLineEpsilonDataSet.Columns.REMARKS.Name, typeof(string));
            ds.Tables[0].Columns.Add(SalesOrderLineEpsilonDataSet.Columns.CORRECTION_VAT_CODE.Name, typeof(string));

            return ds;

        }

        ///// <summary>
        ///// Executes the print events.
        ///// </summary>
        ///// <param name="conn">The conn.</param>
        ///// <param name="parameters">The parameters.</param>
        ///// <returns></returns>
        //private static int ExecutePrintEvents(PmxDbConnection conn, object[] parameters)
        //{
        //    PmxPrintReportProvider prov = new PmxPrintReportProvider(conn);

        //    return prov.PrintReportsForEvent((PmxPrintReportEventType)parameters[0], (string)parameters[1], (string)parameters[2], (int)parameters[3]);
        //}

        /// <summary>
        /// Executes the get item info.
        /// </summary>
        /// <param name="conn">The conn.</param>
        /// <param name="parameters">The parameters.</param>
        /// <returns></returns>
        private static PmxItemInfo ExecuteGetItemInfo(PmxDbConnection conn, object[] parameters)
        {
            PmxItemAllConnectionsProvider prov = new PmxItemAllConnectionsProvider(conn);
            return prov.GetCachedItemInfo((string)parameters[0]);
        }

        /// <summary>
        /// Executes the get item price.
        /// </summary>
        /// <param name="conn">The conn.</param>
        /// <param name="parameters">The parameters.</param>
        /// <returns></returns>
        private static ItemPrice ExecuteGetItemPrice(PmxDbConnection conn, object[] parameters)
        {
            PmxItemProvider prov = new PmxItemProvider((SboConnection)conn);
            return prov.GetItemPrice((string)parameters[0], (string)parameters[1], Convert.ToDouble(parameters[2]), (DateTime)parameters[3]);
        }

        /// <summary>
        /// Executes the make sales order.
        /// </summary>
        /// <param name="conn">The conn.</param>
        /// <param name="parameters">The parameters.</param>
        /// <returns></returns>
        private static int? ExecuteMakeCreditNote(PmxDbConnection conn, object[] parameters)
        {
            int? newDocEntry = null;

            PmxSalesCreditNoteProvider prov = new PmxSalesCreditNoteProvider((SboConnection)conn);


            DataSet ds = (DataSet)parameters[2];

            using (PmxSalesCreditNote bo = prov.GetNewBO())
            {
                bo.CardCode = Convert.ToString(parameters[0]);
                bo.DocDueDate = Convert.ToDateTime(parameters[1]);
                bo["U_FatSer"] = Convert.ToString(parameters[3]);
                bo["U_FatNum"] = Convert.ToString(parameters[4]);
                // bo.Comments = Convert.ToString(parameters[3]);


                string remarks = "";


                foreach (DataRow row in ds.Tables[0].Rows)
                {
                    PmxSalesCreditNoteLine newLine = prov.GetNewAddedDocumentLine(bo);


                    newLine.ItemCode = Convert.ToString(row[SalesOrderLineEpsilonDataSet.Columns.ITEMCODE.Name]);

                    // newLine.UnitPrice = Convert.ToDouble(row[SalesOrderLineEpsilonDataSet.Columns.PRICE.Name]);
                    newLine.Quantity = Convert.ToDouble(row[SalesOrderLineEpsilonDataSet.Columns.QUANTITY.Name]);
                    newLine.Currency = Convert.ToString(row[SalesOrderLineEpsilonDataSet.Columns.CURRENCY.Name]);
                    newLine[Produmex.Sbo.Logex.Data.BusinessObjects.Definitions.Tables.PmxDocumentLineColumns.StorageLocationCode.NAME] = "A001";
                    newLine[Produmex.Sbo.Logex.Data.BusinessObjects.Definitions.Tables.PmxDocumentLineColumns.QualityStatusCode.NAME] = "RELEASED";

                    string tempRemarks = Convert.ToString(row[SalesOrderLineEpsilonDataSet.Columns.REMARKS.Name]);

                    if (!string.IsNullOrEmpty(tempRemarks))
                    {
                        if (!string.IsNullOrEmpty(remarks))
                        {
                            remarks += Environment.NewLine;
                        }
                        remarks += tempRemarks;
                    }

                    string correctionVAT = Convert.ToString(row[SalesOrderLineEpsilonDataSet.Columns.CORRECTION_VAT_CODE.Name]);

                    if (!string.IsNullOrEmpty(correctionVAT))
                    {
                        newLine.VatGroup = correctionVAT;
                    }


                }

                if (!string.IsNullOrEmpty(remarks))
                {
                    bo.Comments += remarks;
                }

                newDocEntry = Convert.ToInt32(prov.AddBO(bo));
            }

            return newDocEntry;
        }

        /// <summary>
        /// Executes the get sbo company info.
        /// </summary>
        /// <param name="conn">The conn.</param>
        /// <param name="parameters">The parameters.</param>
        /// <returns></returns>
        private static SboCompanyAdminInfo ExecuteGetSboCompanyInfo(PmxDbConnection conn, object[] parameters)
        {
            SboCompanyProvider prov = new SboCompanyProvider((SboConnection)conn);
            return prov.GetAdminInfo();
        }

        private static Dictionary<string, PmxParameterValue> ExecuteGetParametersForFlow(PmxDbConnection conn, object[] parameters)
        {
            //get the parameters for this flow
            PmxParameterValueProvider parProv = new PmxParameterValueProvider(conn);
            return parProv.GetParameterValuesAsDictionary(String.Format("{0}->{1}", (string)parameters[0], PmxPropertyCodes.OSE_PARAMETERSET), true, false);

        }


        /// <summary>
        /// Gets the number of rows.
        /// </summary>
        /// <param name="ds">The ds.</param>
        /// <returns></returns>
        private int GetNumberOfRows(DataSet ds)
        {
            if (ds != null && ds.Tables.Count > 0)
            {
                return ds.Tables[0].Rows.Count;
            }
            return 0;
        }
        #endregion

        #region QUERY

        /// <summary>
        /// Static class grouping our native queries
        /// </summary>
        private class BuildQuery
        {
            /// <summary>
            /// Gets the name of the products from stock order by.
            /// </summary>
            /// <param name="gtin">The gtin.</param>
            /// <param name="itemCode">The item code.</param>
            /// <param name="packagingUnit">The packaging unit.</param>
            /// <param name="cardCode">The card code.</param>
            /// <param name="columnNameNumInSales">The column name num in sales.</param>
            /// <returns></returns>
            public static string GetProductsAndSpecialPrices(string gtin, string itemCode, Produmex.Sbo.Logex.Data.SboCommands.PackagingUnit packagingUnit, string cardCode, string columnNameNumInSales, bool filterOutFrozenItems, DbTool dbTool)
            {
                DbQueryBuilder strBuilder = dbTool.GetQueryBuilder();
                strBuilder.Append("SELECT DISTINCT ");

                SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, SboItemTable.NAME, SboItemTable.Columns.ItemCode.NAME, ProductDataSet.Columns.PRODUCT_CODE.Name);

                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, SboItemTable.NAME, PmxItemTable.Columns.CustomItemDescr.NAME, ProductDataSet.Columns.PRODUCT_DESCRIPTION.Name);

                strBuilder.Append(", ");
                strBuilder.Append(SboRecordsetHelper.GetSelectPartsColumnAndIsColumnNull(PmxItemTable.NAME, PmxItemTable.Columns.DefaultStorageLocationOrZoneCode.NAME, ProductDataSet.Columns.DEFAULT_LOCATION_CODE.Name));

                strBuilder.Append(", ");
                Produmex.Sbo.Logex.Data.SboCommands.AddBarcodeUsingPackagingUnit(strBuilder, packagingUnit, ProductDataSet.Columns.GTIN.Name);

                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, SboItemTable.NAME, PmxItemTable.Columns.HasBestBeforeDate.NAME, ProductDataSet.Columns.HAS_BEST_BEFORE.Name);

                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnNameWithAs(strBuilder, SboItemTable.NAME, SboItemTable.Columns.ManBtchNum.NAME, ProductDataSet.Columns.MANBTCHNUM.Name);

                strBuilder.Append(", ");
                SqlCommandHelper.AddTableColumnName(strBuilder, SboItemTable.NAME, PmxItemTable.Columns.IsLogisticUnit.NAME);

                strBuilder.Append(", ", strBuilder.AsIdentifier(SboItemTable.NAME, SboItemTable.Columns.InventoryUom.NAME));


                strBuilder.Append(", ", strBuilder.AsIdentifier(SboItemTable.NAME, SboItemTable.Columns.ItemCost.NAME));
                strBuilder.Append(", ", strBuilder.AsIdentifier(SboItemTable.NAME, columnNameNumInSales));

                //  strBuilder.Append(", ", strBuilder.AsIdentifier(SboSpecialPricesTable.NAME, SboSpecialPricesTable.Columns.Price.NAME));
                //  strBuilder.Append(", ", strBuilder.AsIdentifier(SboSpecialPricesTable.NAME, SboSpecialPricesTable.Columns.Currency.NAME));

                strBuilder.AppendFormat(", ( " + strBuilder.IFNULL + "(SUM({0}),0) - " + strBuilder.IFNULL + "(SUM({1}),0)) AS {2} ",
                    SqlCommandHelper.EscapeTableColumnName(PmxInventoryTotalTable.NAME, PmxInventoryTotalTable.Columns.Quantity.NAME),
                    SqlCommandHelper.EscapeTableColumnName(PmxInventoryLockingDetailTable.NAME, PmxInventoryLockingDetailTable.Columns.Quantity.NAME),
                    SqlCommandHelper.EscapeColumnName(PmxInventoryTotalTable.Columns.Quantity.NAME));



                //strBuilder.Append(", ");
                //SqlCommandHelper.AddColumnNameWithAs(strBuilder, "VatCrctn", "VatCrctn");


                strBuilder.Append(" FROM ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(SboItemTable.NAME));

                //   strBuilder.Append(" LEFT JOIN ");
                //   strBuilder.Append(SqlCommandHelper.EscapeTableName(SboSpecialPricesTable.NAME));
                //   strBuilder.Append(" ON ", strBuilder.AsIdentifier(SboSpecialPricesTable.NAME, SboSpecialPricesTable.Columns.ItemCode.NAME));
                //   strBuilder.Append(" = ", strBuilder.AsIdentifier(SboItemTable.NAME, SboItemTable.Columns.ItemCode.NAME));

                //   strBuilder.Append(" AND ", strBuilder.AsIdentifier(SboSpecialPricesTable.NAME, SboSpecialPricesTable.Columns.CardCode.NAME));
                //   strBuilder.Append(" = ");
                //   strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(cardCode, CultureInfo.InvariantCulture));


                strBuilder.Append(" LEFT JOIN ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxInventoryTotalTable.NAME));
                strBuilder.Append(" ON ", strBuilder.AsIdentifier(PmxInventoryTotalTable.NAME, PmxInventoryTotalTable.Columns.ItemCode.NAME));
                strBuilder.Append(" = ", strBuilder.AsIdentifier(SboItemTable.NAME, SboItemTable.Columns.ItemCode.NAME));


                strBuilder.Append(" LEFT JOIN ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxInventoryLockingDetailTable.NAME));
                strBuilder.Append(" ON ", strBuilder.AsIdentifier(PmxInventoryLockingDetailTable.NAME, PmxInventoryLockingDetailTable.Columns.ItemCode.NAME));
                strBuilder.Append(" = ", strBuilder.AsIdentifier(SboItemTable.NAME, SboItemTable.Columns.ItemCode.NAME));

                strBuilder.Append(" LEFT JOIN ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(PmxItemPackagingTypeTable.NAME));
                strBuilder.Append(" ON ", strBuilder.AsIdentifier(PmxItemPackagingTypeTable.NAME, PmxItemPackagingTypeTable.Columns.ItemCode.NAME));
                strBuilder.Append(" = ", strBuilder.AsIdentifier(SboItemTable.NAME, SboItemTable.Columns.ItemCode.NAME));


                //strBuilder.Append(" LEFT JOIN ");
                //strBuilder.Append(SqlCommandHelper.EscapeTableName(SboVatGroupTable.NAME));
                //strBuilder.Append(" ON ");
                //SqlCommandHelper.AddTableColumnName(strBuilder, SboVatGroupTable.NAME, SboVatGroupTable.Columns.Code.NAME);
                //strBuilder.Append(" = ");
                //SqlCommandHelper.AddTableColumnName(strBuilder, SboItemTable.NAME, SboItemTable.Columns.VatGroupSales.NAME);




                string operand = " WHERE ";
                if (!string.IsNullOrEmpty(gtin))
                {
                    strBuilder.Append(operand);
                    Produmex.Sbo.Logex.Data.SboCommands.AddWhereClauseSearchForBarcode(strBuilder, gtin, true);
                    operand = " AND ";
                }

                if (!string.IsNullOrEmpty(itemCode))
                {
                    strBuilder.Append(operand);
                    strBuilder.Append(SqlCommandHelper.EscapeTableName(SboItemTable.NAME));
                    strBuilder.Append(".");
                    strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboItemTable.Columns.ItemCode.NAME));
                    strBuilder.Append(" = ");
                    strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(itemCode, CultureInfo.InvariantCulture));
                    operand = " AND ";
                }

                //strBuilder.Append(operand);
                //strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboItemTable.NAME));
                //strBuilder.Append(".");
                //strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboItemTable.Columns.InventoryItem.NAME));
                //strBuilder.Append(" = ");
                //strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(SboBooleanColumnValues.Y.VALUE, CultureInfo.InvariantCulture));


                // strBuilder.Append(" AND ");
                //strBuilder.Append(SqlCommandHelper.EscapeTableName(SboItemTable.NAME));
                //strBuilder.Append(".");
                //strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboItemTable.Columns.SalesItem.NAME));
                //strBuilder.Append(" = ");
                //strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(SboBooleanColumnValues.Y.VALUE, CultureInfo.InvariantCulture));


                if (filterOutFrozenItems)
                {
                    Produmex.Sbo.Logex.Data.SboCommands.GetFrozenItemFilterSQLWherePart(strBuilder, operand);
                }


                strBuilder.Append(" GROUP BY ", strBuilder.AsIdentifier(SboItemTable.NAME, SboItemTable.Columns.ItemCode.NAME));

                strBuilder.Append(", ", strBuilder.AsIdentifier(SboItemTable.NAME, PmxItemTable.Columns.CustomItemDescr.NAME));


                strBuilder.Append(", ", strBuilder.AsIdentifier(SboItemTable.NAME, PmxItemTable.Columns.DefaultStorageLocationOrZoneCode.NAME));

                strBuilder.Append(", ", strBuilder.AsIdentifier(SboItemTable.NAME, PmxItemTable.Columns.CodeBars.NAME));
                strBuilder.Append(", ", strBuilder.AsIdentifier(SboItemTable.NAME, PmxItemTable.Columns.SalesBarcode.NAME));
                strBuilder.Append(", ", strBuilder.AsIdentifier(SboItemTable.NAME, PmxItemTable.Columns.NumberOfItemsPerSalesUnit.NAME));
                strBuilder.Append(", ", strBuilder.AsIdentifier(SboItemTable.NAME, PmxItemTable.Columns.PurchaseBarcode.NAME));
                strBuilder.Append(", ", strBuilder.AsIdentifier(SboItemTable.NAME, SboItemTable.Columns.NumberOfItemsPerPurchaseUnit.NAME));

                strBuilder.Append(", ", strBuilder.AsIdentifier(SboItemTable.NAME, PmxItemTable.Columns.HasBestBeforeDate.NAME));

                strBuilder.Append(", ", strBuilder.AsIdentifier(SboItemTable.NAME, PmxItemTable.Columns.ManBtchNum.NAME));

                strBuilder.Append(", ", strBuilder.AsIdentifier(SboItemTable.NAME, PmxItemTable.Columns.IsLogisticUnit.NAME));

                strBuilder.Append(", ", strBuilder.AsIdentifier(SboItemTable.NAME, SboItemTable.Columns.InventoryUom.NAME));

                strBuilder.Append(", ", strBuilder.AsIdentifier(SboItemTable.NAME, SboItemTable.Columns.ItemCost.NAME));

                //strBuilder.Append(", ", strBuilder.AsIdentifier(SboSpecialPricesTable.NAME, SboSpecialPricesTable.Columns.Price.NAME));

                //strBuilder.Append(", ", strBuilder.AsIdentifier(SboSpecialPricesTable.NAME, SboSpecialPricesTable.Columns.Currency.NAME));

                strBuilder.Append(", ", strBuilder.AsIdentifier(SboItemTable.NAME, columnNameNumInSales));
                // strBuilder.Append(", ", strBuilder.AsIdentifier(SboVatGroupTable.NAME, "VatCrctn"));


                strBuilder.Append(" ORDER BY ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(SboItemTable.NAME));
                strBuilder.Append(".");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(PmxItemTable.Columns.CustomItemDescr.NAME));
                return strBuilder.ToString();
            }


            /// <summary>
            /// Gets the customers.
            /// </summary>
            /// <returns></returns>
            public static string GetCustomers()
            {
                StringBuilder strBuilder = new StringBuilder();
                strBuilder.Append("SELECT ");
                SqlCommandHelper.AddColumnNameWithAs(strBuilder, SboBusinessPartnerTable.Columns.CardName.NAME, CustomerDataSet.Columns.CUSTOMER_NAME.Name);

                strBuilder.Append(", ");
                SqlCommandHelper.AddColumnNameWithAs(strBuilder, SboBusinessPartnerTable.Columns.CardCode.NAME, CustomerDataSet.Columns.CUSTOMER_CODE.Name);




                strBuilder.Append(", ");
                SqlCommandHelper.AddColumnNameWithAs(strBuilder, SboBusinessPartnerTable.Columns.Currency.NAME, SboBusinessPartnerTable.Columns.Currency.NAME);

                strBuilder.Append(", ");
                SqlCommandHelper.AddColumnNameWithAs(strBuilder, "VatCrctn", "VatCrctn");


                strBuilder.Append(" FROM ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(SboBusinessPartnerTable.NAME));


                //LEFT OUTER JOIN OVTG ON OCRD.ECVatGroup = OVTG.Code 

                strBuilder.Append(" LEFT JOIN ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(SboVatGroupTable.NAME));
                strBuilder.Append(" ON ");
                SqlCommandHelper.AddTableColumnName(strBuilder, SboVatGroupTable.NAME, SboVatGroupTable.Columns.Code.NAME);
                strBuilder.Append(" = ");
                SqlCommandHelper.AddTableColumnName(strBuilder, SboBusinessPartnerTable.NAME, SboBusinessPartnerTable.Columns.ECVatGroup.NAME);


                strBuilder.Append(" WHERE ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboBusinessPartnerTable.Columns.CardType.NAME));
                strBuilder.Append(" = ");
                strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(SboBusinessPartnerTypeColumnValues.Customer.VALUE, CultureInfo.InvariantCulture));

                GetFrozenCardFilterSQLWherePart(strBuilder, " AND ", CultureInfo.InvariantCulture);

                strBuilder.Append(" ORDER BY ");
                strBuilder.Append(SqlCommandHelper.EscapeColumnName(SboBusinessPartnerTable.Columns.CardName.NAME));
                return strBuilder.ToString();

            }


            /// <summary>
            /// Gets the maximum quantity for credit note.
            /// </summary>
            /// <param name="cardCode">The card code.</param>
            /// <param name="itemCode">The item code.</param>
            /// <param name="dbTool">The database tool.</param>
            /// <returns></returns>
            public static string GetMaxQuantityForCreditNote(string cardCode, string itemCode, DbTool dbTool)
            {

                /*select ISNULL( SUM(INV1.InvQty) - ISNULL( credit.CreditQty, 0), 0) as MaxQuantity
from OINV 
inner join INV1
on inv1.DocEntry = oinv.DocEntry

left join
(
select SUM(RIN1.InvQty) as CreditQty , CardCode, ItemCode
from ORIN 
inner join RIN1
on RIN1.DocEntry = oRIN.DocEntry
where oRIN.CardCode = 'C20000' 
and ItemCode = '00001'
group by CardCode, ItemCode
) as Credit
on credit.CardCode = oinv.CardCode
and inv1.ItemCode = inv1.ItemCode
where oinv.CardCode = 'C20000' 
and inv1.ItemCode = '00001'
group by credit.CreditQty*/

                string creditTableAlias = "Credit";

                DbQueryBuilder strBuilder = dbTool.GetQueryBuilder();

                strBuilder.AppendFormat("SELECT ISNULL( SUM({0}) - ISNULL( {1}, 0), 0) ",
                SqlCommandHelper.EscapeTableColumnName(SboSalesInvoiceLineTable.NAME, SboSalesInvoiceLineTable.Columns.InventoryQuantity.NAME),
                SqlCommandHelper.EscapeTableColumnName(creditTableAlias, SboSalesCreditNoteLineTable.Columns.InventoryQuantity.NAME));

                strBuilder.Append(" FROM ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(SboSalesInvoiceTable.NAME));

                strBuilder.Append(" LEFT JOIN ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(SboSalesInvoiceLineTable.NAME));
                strBuilder.Append(" ON ", strBuilder.AsIdentifier(SboSalesInvoiceTable.NAME, SboSalesInvoiceTable.Columns.DocEntry.NAME));
                strBuilder.Append(" = ", strBuilder.AsIdentifier(SboSalesInvoiceLineTable.NAME, SboSalesInvoiceLineTable.Columns.DocEntry.NAME));


                strBuilder.Append(" LEFT JOIN (");
                {

                    strBuilder.AppendFormat("SELECT SUM({0}) AS {1}, {2}, {3} ",
                        SqlCommandHelper.EscapeTableColumnName(SboSalesCreditNoteLineTable.NAME, SboSalesCreditNoteLineTable.Columns.InventoryQuantity.NAME),
                        SqlCommandHelper.EscapeColumnName(SboSalesCreditNoteLineTable.Columns.InventoryQuantity.NAME),
                        SqlCommandHelper.EscapeTableColumnName(SboSalesCreditNoteLineTable.NAME, SboSalesCreditNoteLineTable.Columns.ItemCode.NAME),
                        SqlCommandHelper.EscapeTableColumnName(SboSalesCreditNoteTable.NAME, SboSalesCreditNoteTable.Columns.CardCode.NAME));

                    strBuilder.Append(" FROM ");
                    strBuilder.Append(SqlCommandHelper.EscapeTableName(SboSalesCreditNoteTable.NAME));

                    strBuilder.Append(" LEFT JOIN ");
                    strBuilder.Append(SqlCommandHelper.EscapeTableName(SboSalesCreditNoteLineTable.NAME));
                    strBuilder.Append(" ON ", strBuilder.AsIdentifier(SboSalesCreditNoteTable.NAME, SboSalesCreditNoteTable.Columns.DocEntry.NAME));
                    strBuilder.Append(" = ", strBuilder.AsIdentifier(SboSalesCreditNoteLineTable.NAME, SboSalesCreditNoteLineTable.Columns.DocEntry.NAME));


                    strBuilder.Append(" WHERE ");
                    strBuilder.Append(SqlCommandHelper.EscapeTableColumnName(SboSalesCreditNoteTable.NAME, SboSalesCreditNoteTable.Columns.CardCode.NAME));
                    strBuilder.Append(" = ");
                    strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(cardCode, CultureInfo.InvariantCulture));

                    strBuilder.Append(" AND ");
                    strBuilder.Append(SqlCommandHelper.EscapeTableColumnName(SboSalesCreditNoteLineTable.NAME, SboSalesCreditNoteLineTable.Columns.ItemCode.NAME));
                    strBuilder.Append(" = ");
                    strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(itemCode, CultureInfo.InvariantCulture));

                    strBuilder.AppendFormat(" GROUP BY {0}, {1}, {2}",
                    SqlCommandHelper.EscapeTableColumnName(SboSalesCreditNoteLineTable.NAME, SboSalesCreditNoteLineTable.Columns.InventoryQuantity.NAME),
                    SqlCommandHelper.EscapeTableColumnName(SboSalesCreditNoteLineTable.NAME, SboSalesCreditNoteLineTable.Columns.ItemCode.NAME),
                        SqlCommandHelper.EscapeTableColumnName(SboSalesCreditNoteTable.NAME, SboSalesCreditNoteTable.Columns.CardCode.NAME));



                }
                strBuilder.Append(" ) AS ");
                strBuilder.Append(SqlCommandHelper.EscapeTableName(creditTableAlias));
                strBuilder.Append(" ON ", strBuilder.AsIdentifier(SboSalesInvoiceTable.NAME, SboSalesInvoiceTable.Columns.CardCode.NAME));
                strBuilder.Append(" = ", strBuilder.AsIdentifier(creditTableAlias, SboSalesCreditNoteTable.Columns.CardCode.NAME));
                strBuilder.Append(" AND ", strBuilder.AsIdentifier(SboSalesInvoiceLineTable.NAME, SboSalesInvoiceLineTable.Columns.ItemCode.NAME));
                strBuilder.Append(" = ", strBuilder.AsIdentifier(creditTableAlias, SboSalesCreditNoteLineTable.Columns.ItemCode.NAME));



                strBuilder.Append(" WHERE ");
                strBuilder.Append(SqlCommandHelper.EscapeTableColumnName(SboSalesInvoiceTable.NAME, SboSalesInvoiceTable.Columns.CardCode.NAME));
                strBuilder.Append(" = ");
                strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(cardCode, CultureInfo.InvariantCulture));

                strBuilder.Append(" AND ");
                strBuilder.Append(SqlCommandHelper.EscapeTableColumnName(SboSalesInvoiceLineTable.NAME, SboSalesInvoiceLineTable.Columns.ItemCode.NAME));
                strBuilder.Append(" = ");
                strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(itemCode, CultureInfo.InvariantCulture));

                strBuilder.Append(" GROUP BY ");
                strBuilder.Append(SqlCommandHelper.EscapeTableColumnName(creditTableAlias, SboSalesCreditNoteLineTable.Columns.InventoryQuantity.NAME));





                return strBuilder.ToString();

            }

            /// <summary>
            /// Gets the frozen item filter SQL where part.
            /// </summary>
            /// <param name="strBuilder">The STR builder.</param>
            /// <param name="operand">The operand.</param>
            /// <param name="cultureDatabase">The culture database.</param>
            public static void GetFrozenCardFilterSQLWherePart(StringBuilder strBuilder, string operand, CultureInfo cultureDatabase)
            {
                Produmex.Foundation.Validate.IsNotNull(strBuilder, "StringBuilder");

                //frozenFor = 'N'
                //OR (frozenFor = 'Y'
                //AND getdate() not between frozenFrom AND frozenTo)

                strBuilder.Append(operand);
                strBuilder.Append(" ( ");
                SqlCommandHelper.AddTableColumnName(strBuilder, SboBusinessPartnerTable.NAME, SboBusinessPartnerTable.Columns.frozenFor.NAME);
                strBuilder.Append(" = ");
                strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(SboBooleanColumnValues.N.VALUE, cultureDatabase));
                strBuilder.Append(" OR (");
                SqlCommandHelper.AddTableColumnName(strBuilder, SboBusinessPartnerTable.NAME, SboBusinessPartnerTable.Columns.frozenFor.NAME);
                strBuilder.Append(" = ");
                strBuilder.Append(SqlCommandHelper.FormatAndEscapeColumnValue(SboBooleanColumnValues.Y.VALUE, cultureDatabase));
                strBuilder.Append(" AND GetDate() NOT BETWEEN ");
                SqlCommandHelper.AddTableColumnName(strBuilder, SboBusinessPartnerTable.NAME, SboBusinessPartnerTable.Columns.frozenFrom.NAME);
                strBuilder.Append(" AND ");
                SqlCommandHelper.AddTableColumnName(strBuilder, SboBusinessPartnerTable.NAME, SboBusinessPartnerTable.Columns.frozenTo.NAME);
                strBuilder.Append(" )) ");

            }

        }
        #endregion
    }
}
